@@ -3,6 +3,9 @@
  * Copyright 2014-2019 devemux86
  * Copyright 2017 usrusr
  * Copyright 2019 cpt1gl0
+ * Copyright 2019 Adrian Batzill
+ * Copyright 2019 Matthew Egeler
+ * Copyright 2019 mg4gh
  *
  * This program is free software: you can redistribute it and/or modify it under the
  * terms of the GNU Lesser General Public License as published by the Free Software
@@ -17,32 +20,21 @@
  */
 package org.mapsforge.map.awt.graphics;
 
-import org.mapsforge.core.graphics.Bitmap;
 import org.mapsforge.core.graphics.Canvas;
 import org.mapsforge.core.graphics.Color;
-import org.mapsforge.core.graphics.Filter;
-import org.mapsforge.core.graphics.Matrix;
 import org.mapsforge.core.graphics.Paint;
-import org.mapsforge.core.graphics.Path;
-import org.mapsforge.core.graphics.Style;
+import org.mapsforge.core.graphics.*;
 import org.mapsforge.core.model.Dimension;
 import org.mapsforge.core.model.Rectangle;
+import org.mapsforge.core.util.Parameters;
 
-import java.awt.AlphaComposite;
-import java.awt.Composite;
-import java.awt.Graphics2D;
-import java.awt.RenderingHints;
+import java.awt.*;
 import java.awt.color.ColorSpace;
 import java.awt.font.TextLayout;
 import java.awt.geom.AffineTransform;
 import java.awt.geom.Area;
 import java.awt.geom.Rectangle2D;
-import java.awt.image.BufferedImage;
-import java.awt.image.BufferedImageOp;
-import java.awt.image.ColorConvertOp;
-import java.awt.image.IndexColorModel;
-import java.awt.image.LookupOp;
-import java.awt.image.ShortLookupTable;
+import java.awt.image.*;
 import java.util.AbstractMap;
 import java.util.Map;
 
@@ -94,7 +86,7 @@ class AwtCanvas implements Canvas {
 
     AwtCanvas(Graphics2D graphics2D) {
         this.graphics2D = graphics2D;
-        enableAntiAliasing();
+        setAntiAlias(Parameters.ANTI_ALIASING);
 
         createFilters();
     }
@@ -167,8 +159,15 @@ class AwtCanvas implements Canvas {
     }
 
     @Override
-    public void drawBitmap(Bitmap bitmap, int left, int top, Filter filter) {
+    public void drawBitmap(Bitmap bitmap, int left, int top, float alpha, Filter filter) {
+        Composite composite = this.graphics2D.getComposite();
+        if (alpha != 1) {
+            this.graphics2D.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, alpha));
+        }
         this.graphics2D.drawImage(applyFilter(AwtGraphicFactory.getBitmap(bitmap), filter), left, top, null);
+        if (alpha != 1) {
+            this.graphics2D.setComposite(composite);
+        }
     }
 
     @Override
@@ -178,8 +177,15 @@ class AwtCanvas implements Canvas {
     }
 
     @Override
-    public void drawBitmap(Bitmap bitmap, Matrix matrix, Filter filter) {
+    public void drawBitmap(Bitmap bitmap, Matrix matrix, float alpha, Filter filter) {
+        Composite composite = this.graphics2D.getComposite();
+        if (alpha != 1) {
+            this.graphics2D.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, alpha));
+        }
         this.graphics2D.drawRenderedImage(applyFilter(AwtGraphicFactory.getBitmap(bitmap), filter), AwtGraphicFactory.getAffineTransform(matrix));
+        if (alpha != 1) {
+            this.graphics2D.setComposite(composite);
+        }
     }
 
     @Override
@@ -193,11 +199,18 @@ class AwtCanvas implements Canvas {
 
     @Override
     public void drawBitmap(Bitmap bitmap, int srcLeft, int srcTop, int srcRight, int srcBottom,
-                           int dstLeft, int dstTop, int dstRight, int dstBottom, Filter filter) {
+                           int dstLeft, int dstTop, int dstRight, int dstBottom, float alpha, Filter filter) {
+        Composite composite = this.graphics2D.getComposite();
+        if (alpha != 1) {
+            this.graphics2D.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, alpha));
+        }
         this.graphics2D.drawImage(applyFilter(AwtGraphicFactory.getBitmap(bitmap), filter),
                 dstLeft, dstTop, dstRight, dstBottom,
                 srcLeft, srcTop, srcRight, srcBottom,
                 null);
+        if (alpha != 1) {
+            this.graphics2D.setComposite(composite);
+        }
     }
 
     @Override
@@ -260,6 +273,40 @@ class AwtCanvas implements Canvas {
         throw new IllegalArgumentException(UNKNOWN_STYLE + style);
     }
 
+    @Override
+    public void drawPathText(String text, Path path, Paint paint) {
+        if (text == null || text.trim().isEmpty()) {
+            return;
+        }
+        if (paint.isTransparent()) {
+            return;
+        }
+
+        AwtPaint awtPaint = AwtGraphicFactory.getPaint(paint);
+        AwtPath awtPath = AwtGraphicFactory.getPath(path);
+
+        if (awtPaint.stroke == null) {
+            this.graphics2D.setColor(awtPaint.color);
+            this.graphics2D.setFont(awtPaint.font);
+        } else {
+            setColorAndStroke(awtPaint);
+        }
+
+        TextStroke textStroke = new TextStroke(text, awtPaint.font, this.graphics2D.getFontRenderContext(), false, false);
+        Style style = awtPaint.style;
+        switch (style) {
+            case FILL:
+                this.graphics2D.fill(textStroke.createStrokedShape(awtPath.path2D));
+                return;
+
+            case STROKE:
+                this.graphics2D.draw(textStroke.createStrokedShape(awtPath.path2D));
+                return;
+        }
+
+        throw new IllegalArgumentException(UNKNOWN_STYLE + style);
+    }
+
     @Override
     public void drawText(String text, int x, int y, Paint paint) {
         if (text == null || text.trim().isEmpty()) {
@@ -355,6 +402,8 @@ class AwtCanvas implements Canvas {
     @Override
     public void setAntiAlias(boolean aa) {
         this.graphics2D.setRenderingHint(RenderingHints.KEY_ANTIALIASING, aa ? RenderingHints.VALUE_ANTIALIAS_ON : RenderingHints.VALUE_ANTIALIAS_OFF);
+        this.graphics2D.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, aa ? RenderingHints.VALUE_TEXT_ANTIALIAS_ON : RenderingHints.VALUE_TEXT_ANTIALIAS_OFF);
+        this.graphics2D.setRenderingHint(RenderingHints.KEY_FRACTIONALMETRICS, aa ? RenderingHints.VALUE_FRACTIONALMETRICS_ON : RenderingHints.VALUE_FRACTIONALMETRICS_OFF);
     }
 
     @Override
@@ -365,7 +414,7 @@ class AwtCanvas implements Canvas {
         } else {
             this.bufferedImage = AwtGraphicFactory.getBitmap(bitmap);
             this.graphics2D = this.bufferedImage.createGraphics();
-            enableAntiAliasing();
+            setAntiAlias(Parameters.ANTI_ALIASING);
             this.graphics2D.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);
             this.graphics2D.setRenderingHint(RenderingHints.KEY_STROKE_CONTROL, RenderingHints.VALUE_STROKE_PURE);
         }
@@ -373,6 +422,11 @@ class AwtCanvas implements Canvas {
 
     @Override
     public void setClip(int left, int top, int width, int height) {
+        setClip(left, top, width, height, false);
+    }
+
+    @Override
+    public void setClip(int left, int top, int width, int height, boolean intersect) {
         this.graphics2D.setClip(left, top, width, height);
     }
 
@@ -430,12 +484,6 @@ class AwtCanvas implements Canvas {
         this.graphics2D.setComposite(oldComposite);
     }
 
-    private void enableAntiAliasing() {
-        this.graphics2D.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
-        this.graphics2D.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
-        this.graphics2D.setRenderingHint(RenderingHints.KEY_FRACTIONALMETRICS, RenderingHints.VALUE_FRACTIONALMETRICS_ON);
-    }
-
     private void fillColor(java.awt.Color color) {
         final Composite originalComposite = this.graphics2D.getComposite();
         this.graphics2D.setComposite(AlphaComposite.getInstance(color.getAlpha() == 0 ? AlphaComposite.CLEAR : AlphaComposite.SRC_OVER));
@@ -444,7 +492,7 @@ class AwtCanvas implements Canvas {
         this.graphics2D.setComposite(originalComposite);
     }
 
-    public void setColorAndStroke(AwtPaint awtPaint) {
+    void setColorAndStroke(AwtPaint awtPaint) {
         this.graphics2D.setColor(awtPaint.color);
         if (awtPaint.stroke != null) {
             this.graphics2D.setStroke(awtPaint.stroke);
